/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      XML
Version:   1.0
Status:    In Development
Date:      January 2004
ID:        ID_XML
Include:   pandora/data/xml.h
Category:  Data
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2001-2004.  All rights reserved.
Keywords:  xml, data, text, process, statement, language, markup, extensible

###DESCRIPTION###
<p>The XML class provides the necessary functionality to create and maintain
XML data files.  It is capable of interpreting and validating XML files with
or without correct structure and can perform various data manipulations while
doing so.  The XML class is also designed to minimise the amount of resources
used in storing XML information and exhibits excellent performance in its
processing.</p>

<p>Data can be loaded into an XML object either by specifying a file
<a href="#Location">Location</a> or by giving it an XML
<a href="#Statement">Statement</a>.  If you have multiple XML statements to
process, you can reset the Location or Statement fields after initialisation
and the XML object will rebuild itself.  This saves you from having to
allocate multiple XML objects for batch processing.</p>

<p>Once an XML object has interpreted a statement, you can read the
information by scanning the array stored in the <a href="#Tags">Tags</a>
field.  This array contains an XMLTag structure for each tag found in the
original XML statement.  For more information on how to scan this information,
refer to the <a href="#Tags">Tags</a> field.</p>

###END###

CDATA_SECTION           <ELEMENT><![CDATA[a CDATA section is used to escape a block of text that would otherwise be recognized as markup.]]></ELEMENT>
ENTITY_REFERENCE        <Element>&ent;</Element>
ENTITY                  <!ENTITY % textgizmo "fontgizmo">
PROCESSING_INSTRUCTION  <?XML version="1.0" standalone="yes" ?>
DOCUMENT_TYPE *         <!DOCTYPE INVOICE SYSTEM "http://www.xml.com/product.dtd">
NOTATION                <!NOTATION gif SYSTEM "viewer.exe">

*****************************************************************************/

//#define DEBUG

#include <pandora/system/all.h>
#include <pandora/modules/strings.h>
#include <pandora/files/file.h>
#include <pandora/data/xml.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_XML 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_XML, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2001-2004.  All rights reserved.",
   "January 2004",
   "XML"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static OBJECTPTR XMLClass = 0;
static OBJECTPTR StringsMod = 0;

/*****************************************************************************
** Class definition.
*/

struct XML {
   OBJECT_HEADER            /* Class header */
   STRING Location;         /* Location of the XML data file */
   struct XMLTag **Tags;    /* Array of tag pointers, in linear order.  Useful for looking up indexes */
   LONG   TagCount;         /* Total number of tags loaded into the array */
   LONG   Flags;            /* Optional user flags */
   LONG   CurrentTag;       /* Current Tag - used for certain operations */
   LONG   PrivateDataSize;  /* Extra data can be allocated per tag, according to the number of bytes specified here */
   LARGE  FileSize;         /* The byte-size of the original XML file */
   LARGE  TimeStamp;        /* The time-stamp set against the original XML file */

   PRIVATE_FIELDS
   ERROR prvCompileError;   /* */
   LONG  prvBalance;        /* Indicates that the tag structure is correctly balanced if zero */
};

static ERROR GET_Location(struct XML *, STRING *);
static ERROR GET_Statement(struct XML *, STRING *);

static ERROR SET_Location(struct XML *, STRING);
static ERROR SET_Statement(struct XML *, STRING);

static struct FieldDef XMLFlags[] = {
   { "WELLFORMED",      XMF_WELLFORMED },
   { "INCLUDECOMMENTS", XMF_INCLUDECOMMENTS },
   { "STRIPCONTENT",    XMF_STRIPCONTENT },
   { "LOWERCASE",       XMF_LOWERCASE },
   { "UPPERCASE",       XMF_UPPERCASE },
   { "INDENT",          XMF_INDENT },
   { NULL, NULL }
};

static struct FieldArray XMLFields[] = {
   { "Location",        0, FDF_STRING|FDF_RW,    0, NULL, SET_Location },
   { "Tags",            0, FDF_POINTER|FDF_R,    0, NULL, NULL },
   { "TagCount",        0, FDF_LONG|FDF_R,       0, NULL, NULL },
   { "Flags",           0, FDF_LONGFLAGS|FDF_RW, (LONG)&XMLFlags, NULL, NULL },
   { "CurrentTag",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "PrivateDataSize", 0, FDF_LONG|FDF_RI,      0, NULL, NULL },
   { "FileSize",        0, FDF_LARGE|FDF_R,      0, NULL, NULL },
   { "TimeStamp",       0, FDF_LARGE|FDF_R,      0, NULL, NULL },
   /*** Virtual fields ***/
   { "Src",             0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   { "Statement",       0, FDF_STRING|FDF_RW, 0, GET_Statement, SET_Statement },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR XML_Clear(struct XML *, APTR);
static ERROR XML_DataChannel(struct XML *, struct acDataChannel *);
static ERROR XML_Free(struct XML *, APTR);
static ERROR XML_GetUnlistedField(struct XML *, struct acGetUnlistedField *);
static ERROR XML_Init(struct XML *, APTR);
static ERROR XML_NewObject(struct XML *, APTR);
static ERROR XML_Reset(struct XML *, APTR);
static ERROR XML_SaveToObject(struct XML *, struct acSaveToObject *);

static struct ActionArray XMLActions[] = {
   { AC_Clear,            XML_Clear },
   { AC_DataChannel,      XML_DataChannel },
   { AC_Free,             XML_Free },
   { AC_GetUnlistedField, XML_GetUnlistedField },
   { AC_Init,             XML_Init },
   { AC_NewObject,        XML_NewObject },
   { AC_Reset,            XML_Reset },
   { AC_SaveToObject,     XML_SaveToObject },
   { NULL, NULL }
};

/*****************************************************************************
** Methods
*/

static ERROR XML_AddXMLStatement(struct XML *, struct mtAddXMLStatement *);
static ERROR XML_GetXMLContent(struct XML *, struct mtGetXMLContent *);
static ERROR XML_GetXMLString(struct XML *, struct mtGetXMLString *);
static ERROR XML_RemoveXMLTag(struct XML *, struct mtRemoveXMLTag *);
static ERROR XML_SetXMLAttribute(struct XML *, struct mtSetXMLAttribute *);
static ERROR XML_SortXML(struct XML *, struct mtSortXML *);

static struct FunctionField argsAddXMLStatement[] = { { NULL, NULL } };
static struct FunctionField argsGetXMLContent[]   = { { NULL, NULL } };
static struct FunctionField argsGetXMLString[]    = { { NULL, NULL } };
static struct FunctionField argsRemoveXMLTag[]    = { { "Index", ARG_LONG }, { "Total", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsSetXMLAttribute[] = { { "Index", ARG_LONG }, { "Attrib", ARG_LONG }, { "Name", ARG_STR }, { "Value", ARG_STR }, { NULL, NULL } };
static struct FunctionField argsSortXML[]         = { { "Index", ARG_LONG }, { "List", ARG_PTR }, { NULL, NULL } };

static struct MethodArray XMLMethods[] = {
   { MT_GetXMLContent,   XML_GetXMLContent,   "GetContent",   argsGetXMLContent,   sizeof(struct mtGetXMLContent) },
   { MT_GetXMLString,    XML_GetXMLString,    "GetString",    argsGetXMLString,    sizeof(struct mtGetXMLString) },
   { MT_SetXMLAttribute, XML_SetXMLAttribute, "SetAttribute", argsSetXMLAttribute, sizeof(struct mtSetXMLAttribute) },
   { MT_AddXMLStatement, XML_AddXMLStatement, "AddStatement", argsAddXMLStatement, sizeof(struct mtAddXMLStatement) },
   { MT_SortXML,         XML_SortXML,         "SortXML",      argsSortXML,         sizeof(struct mtSortXML) },
   { MT_RemoveXMLTag,    XML_RemoveXMLTag,    "RemoveTag",    argsRemoveXMLTag,    sizeof(struct mtRemoveXMLTag) },
   { NULL, NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Prototypes
*/

struct XMLTag * BuildXMLString(struct XMLTag *, STRING, LONG, LONG *);
static ERROR ConvertXML(struct XML *, STRING);
static ERROR ExtractTag(struct XML *, STRING, STRING *, LONG *);
static ERROR ExtractContent(struct XML *, STRING, STRING *, LONG *);
struct XMLTag * LengthXMLString(struct XMLTag *, LONG *);
static void TagCount(struct XMLTag *, LONG *);

#ifdef DEBUG

static void DebugTaglist(struct XML *Self)
{
   WORD i;
   for (i=0; i < Self->TagCount; i++) {
      DPrintF("Tag:","%.3d/%.3d: $%.8x Next: $%.8x, Prev: $%.8x, Child: $%.8x %s", i, Self->Tags[i]->Index, Self->Tags[i], Self->Tags[i]->Next, Self->Tags[i]->Prev, Self->Tags[i]->Child, Self->Tags[i]->Attributes[0].Name);
   }
}

static void DebugTree(struct XMLTag *Tag, LONG *Indent)
{
   struct XMLTag *child;
   LONG i, j;
   UBYTE buffer[1000];

   if (!Indent) {
      i = 0;
      while (Tag) {
         DebugTree(Tag, &i);
         Tag=Tag->Next;
      }
      return;
   }

   if (!Tag->Attributes) {
      DPrintF("!DebugTree:","Error - no attributes in tag %d", Tag->Index);
      return;
   }

   for (i=0; i < *Indent; i++) buffer[i] = ' ';

   if (Tag->Attributes[0].Name) {
      for (j=0; Tag->Attributes[0].Name[j]; j++) buffer[i++] = Tag->Attributes[0].Name[j];
   }
   else {
      /*** Extract up to 20 characters of content ***/
      for (j=0; (Tag->Attributes[0].Value[j]) AND (j < 20); j++) buffer[i++] = Tag->Attributes[0].Value[j];
   }
   buffer[i] = 0;

   DPrintF("Tag:","%s", buffer);

   Indent[0]++;
   for (child=Tag->Child; child; child=child->Next) {
      DebugTree(child, Indent);
   }
   Indent[0]--;
}

#endif

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&XMLClass, NULL,
      FID_BaseClassID|TLONG, ID_XML,
      FID_Version|TFLOAT,    VER_XML,
      FID_Name|TSTRING,      "XML",
      FID_Category|TLONG,    CCF_DATA,
      FID_Actions|TPTR,      XMLActions,
      FID_Methods|TPTR,      XMLMethods,
      FID_Fields|TPTR,       XMLFields,
      FID_Size|TLONG,        sizeof(struct XML),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (StringsMod) { Action(AC_Free, StringsMod, NULL); StringsMod = NULL; }
   if (XMLClass) { Action(AC_Free, XMLClass, NULL); XMLClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###PRIVATE###
Name:  AddXMLStatement
Short: Adds an XML statement to an XML object.

###DESCRIPTION###

###INPUT###
LONG/Index: The index at which the statement should be inserted (zero is the beginning).
STRING/Statement: The statement to process.
*XMLTag/Result: The resulting tag pointer.

###RESULT###
ERR_Okay: The statement was added successfully.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static ERROR XML_AddXMLStatement(struct XML *Self, struct mtAddXMLStatement *Args)
{
   if (!Args) return(ObjectError(Self, MT_AddXMLStatement, ERR_Args));

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Clear
Short: Clears all of the data held in an XML object.
###END###

*****************************************************************************/

static ERROR XML_Clear(struct XML *Self, APTR Void)
{
   LONG i;

   if (Self->Location)  { FreeMemory(Self->Location);  Self->Location  = NULL; }

   if (Self->Tags) {
      for (i=0; i < Self->TagCount; i++) {
         if (Self->Tags[i]) FreeMemory(Self->Tags[i]);
      }

      FreeMemory(Self->Tags);
      Self->Tags = NULL;

      Self->TagCount = 0;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  DataChannel
Short: XML data can be added to an XML object through this action.
###END###

*****************************************************************************/

static ERROR XML_DataChannel(struct XML *Self, struct acDataChannel *Args)
{
   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_XML) AND (Args->Version <= 1)) {
      struct XML xml;
      struct XMLTag *tag;
      LONG i, index;

      /*** If there is no existing data in our object, just add the XML from scratch ***/

      if (Self->TagCount < 1) return(ConvertXML(Self, Args->Buffer));

      ClearMemory(&xml, sizeof(xml));
      xml.PrivateDataSize = Self->PrivateDataSize;
      xml.Flags = Self->Flags;

      ConvertXML(&xml, Args->Buffer);

      /* Increase the size of our XML tag array.  The tag array
      ** is a list of all the XML tags so that the developer
      ** can look up using indexes.
      */

      if (ReallocMemory(Self->Tags, sizeof(APTR) * (Self->TagCount + xml.TagCount + 1), &Self->Tags, NULL) != ERR_Okay) {
         DPrintF("@DataChannel:","[XML] Failed to reallocate tag array.");
         XML_Free(&xml, NULL);
         return(ERH_ReallocMemory);
      }

      /* Correct the end of the chain to correctly link up.  Note that we append to the
      ** last root tag, which is not necessarily the last tag at the end of the array.  For
      ** that reason, we need the loop.
      */

      if ((tag = Self->Tags[0])) {
         while (tag->Next) tag = tag->Next;
         tag->Next = xml.Tags[0];
         xml.Tags[0]->Prev = tag;
      }

      /*** Copy the new tags into the array and set the tag index numbers ***/

      index = Self->TagCount;
      for (i=0; i < xml.TagCount; i++, index++) {
         Self->Tags[index] = xml.Tags[i];
         Self->Tags[index]->Index = index;
      }

      Self->Tags[index] = NULL; /* Terminate the tag array */
      Self->TagCount = index;   /* Set the new tag count */

      FreeMemory(xml.Tags);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** XML: Free
*/

static ERROR XML_Free(struct XML *Self, APTR Void)
{
   LONG i;

   if (Self->Location)  { FreeMemory(Self->Location);  Self->Location  = NULL; }

   if (Self->Tags) {
      for (i=0; i < Self->TagCount; i++) {
         if (Self->Tags[i]) FreeMemory(Self->Tags[i]);
      }

      FreeMemory(Self->Tags);
      Self->Tags = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  GetUnlistedField
Short: Retrieves data from an xml object.
###END###

*****************************************************************************/

static ERROR XML_GetUnlistedField(struct XML *Self, struct acGetUnlistedField *Args)
{
   struct XMLTag *tags;
   LONG i, j, len, count;

   if (!Args) return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_Args));

   if ((!Args->Field) OR (!Args->Buffer) OR (Args->Size < 1)) {
      return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_Args));
   }

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) {
      return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_Failed));
   }

   Args->Buffer[0] = 0;

   if (StrCompare("TagCount", Args->Field, 0, NULL) IS ERR_Okay) {
      if (Args->Field[8] IS '(') {
         /* If a string is specified inside brackets, get a count
         ** of all XML tags that have the specified name.  Child
         ** tags bearing the same name are not included in the count,
         ** e.g. <menu>..<menu/>..</menu> yields a count of 1.
         */

         for (len=0; (Args->Field[9+len] != ')') AND (Args->Field[9+len]); len++);

         {
            UBYTE buffer[len+1];
            for (i=0; i < len; i++) buffer[i] = Args->Field[9+i];
            buffer[i] = 0;

            count = 0;
            tags = Self->Tags[0];
            while (tags) {
               if (StrCompare(buffer, tags->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  count++;
                  tags = tags->Next; /* Skip all children so that we don't include children with the same tag name */
               }
               else tags = Self->Tags[tags->Index+1];
            }
            Args->Buffer[IntToStr(count, Args->Buffer, Args->Size)] = 0;
         }

         return(ERR_Okay);
      }
      else {
         /*** Return a count of all tags in the XML object ***/
         Args->Buffer[IntToStr(Self->TagCount, Args->Buffer, Args->Size)] = 0;
         return(ERR_Okay);
      }
   }
   else if (StrCompare("Tag(", Args->Field, 0, NULL) IS ERR_Okay) {
      UBYTE tagname[40], attribute[40];
      LONG index, pos;

      /* The tag field allows specific XML tags/attributes to be read.
      **
      **   Format:  Tag(tagname, index, attribute)
      **   Example: Tag(menu, 3, menuitem)
      */

      i = 0;
      for (pos=4; (i < sizeof(tagname)-1) AND (Args->Field[pos]) AND (Args->Field[pos] != ')') AND (Args->Field[pos] != ','); pos++) {
         tagname[i++] = Args->Field[pos];
      }
      tagname[i] = 0;

      if (Args->Field[pos] IS ',') pos++;
      while ((Args->Field[pos] > 0) AND (Args->Field[pos] <= 0x20)) pos++;

      /*** Get the index number ***/

      if ((Args->Field[pos] >= '0') AND (Args->Field[pos] <= '9')) {
         index = StrToInt(Args->Field+pos);
      }
      else index = 0;

      while ((Args->Field[pos]) AND (Args->Field[pos] != ')') AND (Args->Field[pos] != ',')) pos++;
      if (Args->Field[pos] IS ',') pos++;
      while ((Args->Field[pos] > 0) AND (Args->Field[pos] <= 0x20)) pos++;

      /*** Get the name of the requested attribute ***/

      i = 0;
      while ((i < sizeof(attribute)) AND (Args->Field[pos]) AND (Args->Field[pos] != ')') AND (Args->Field[pos] != ',')) {
         attribute[i++] = Args->Field[pos++];
      }
      attribute[i] = 0;

      /*** Find the tag and extract the data ***/

      count = -1;
      tags = Self->Tags[0];
      while (tags) {
         if (StrCompare(tagname, tags->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            count++;
            if (count IS index) {
               for (j=0; j < tags->TotalAttributes; j++) {
                  if (StrCompare(attribute, tags->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     StrCopy(tags->Attributes[j].Value, Args->Buffer, Args->Size);
                     return(ERR_Okay);
                  }
               }
               break;
            }
            else {
               tags = tags->Next; /* Skip all children so that we don't include children with the same tag name */
               continue;
            }
         }
         tags = Self->Tags[tags->Index+1];
      }

      DPrintF("@GetField:","[XML:%d] Search failed: %s", Self->Head.UniqueID, Args->Field);
      return(ERR_Search);
   }
   else {
      DPrintF("GetUnlistedField()","[XML:%d] Unsupported field name \"%s\".", Self->Head.UniqueID, Args->Field);
      return(ERR_Failed);
   }

   return(ERR_Failed);
}

/*****************************************************************************

###METHOD###
Name:  GetXMLContent
Short: Extracts the content embedded inside an XML tag.

###DESCRIPTION###
<P>The GetXMLContent method is used to extract the string content from an XML
tag.  It will extract content that is immediately embedded within the XML
tag and will not perform deep analysis of the tag structure.  Take the following structure
for example:</p>

<pre>
  &lt;body&gt;
    Hello
    &lt;bold&gt;my&lt;/bold%gt;
    friend!
  $lt;/body%gt;
</pre>

<p>This will produce the result "Hello friend!" and miss everything encapsulated
within the bold tag.</p>

###INPUT###
LONG/Index: The index of the tag that contains the content that you want to extract.
STRING/Buffer: Pointer to a buffer that will receive the string data.
LONG/Length: The length of the Buffer in bytes.

###RESULT###
ERR_Okay: The content string was successfully extracted.
ERR_Args: Invalid arguments were specified.
ERR_BufferOverflow: The buffer was not large enough to hold the content (the resulting string will be valid but truncated).

###END###

*****************************************************************************/

static ERROR XML_GetXMLContent(struct XML *Self, struct mtGetXMLContent *Args)
{
   struct XMLTag *tag;
   WORD i, j;

   if ((!Args) OR (!Args->Buffer) OR (!Args->Length)) return(ObjectError(Self, MT_GetXMLContent, ERR_Args));
   if ((Args->Index < 0) OR (Args->Index > Self->TagCount)) return(ObjectError(Self, MT_GetXMLContent, ERR_Args));

   Args->Buffer[0] = 0;
   j = 0;
   if ((tag = Self->Tags[Args->Index]->Child)) {
      for (i=0; i < tag->TotalAttributes; i++) {
         if (!tag->Attributes[i].Name) {
            while ((j < Args->Length) AND (tag->Attributes[i].Value[j])) {
               j += StrCopy(tag->Attributes[i].Value, Args->Buffer+j, Args->Length-j);
            }
         }
      }
   }

   if (j >= Args->Length) return(ObjectError(Self, MT_GetXMLContent, ERR_BufferOverflow));

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    GetXMLString
Synonym: GetString
Short:   Retrieves data from an XML object in standard XML string format.

###DESCRIPTION###
<p>Use the GetXMLString method when you want to build XML strings from data
that has been loaded into an XML object.  You can build information from the
entire XML object or from a specific area of the XML tree by setting the
Index parameter.  Currently supported flags are:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">INCLUDESIBLINGS</td><td>Normally when building XML strings, only the tag itself and its children will be written to the XML statement.  If you want to include other tags sitting at the same level in the XML hierarchy, set this flag.</td></tr>
<tr><td width="10%">LOWERCASE</td><td>Set this flag if you want the XML statement to be written entirely in lower-case.</td></tr>
<tr><td width="10%">UPPERCASE</td><td>Set this flag if you want the XML statement to be written entirely in upper-case.</td></tr>
</table>

<p>The XML string that is built by this method will be stored in the Result
parameter.  You are required to free this memory block when you are finished
with it.</p>

###INPUT###
LONG/Index: The index number of the tag that you want to pull data from (set to zero for all XML data).
LONG/Flags: Special flags that affect the construction of the XML string.
STRING/Result: The resulting string is returned in this parameter.

###RESULT###
ERR_Okay: The XML string was retrieved.
ERR_Args: Invalid arguments were specified.
ERR_NoData: No information has been loaded into the XML object.
ERR_AllocMemory: Failed to allocate an XML string for the result.

###END###

*****************************************************************************/

static ERROR XML_GetXMLString(struct XML *Self, struct mtGetXMLString *Args)
{
   struct XMLTag *tag;
   STRING buffer;
   LONG offset, length, size;

   if ((Self->TagCount <= 0) OR (!Self->Tags)) return(ObjectError(Self, MT_GetXMLString, ERR_FieldNotSet));
   if (!Args) return(ObjectError(Self, MT_GetXMLString, ERR_Args));

   #ifdef DEBUG
      DPrintF("GetXMLString()","Index: %d", Args->Index);
   #endif

   /*** Calculate the size of the buffer required to save the XML information ***/

   if (Args->Flags & XMF_INCLUDESIBLINGS) {
      size = 0;
      tag = Self->Tags[Args->Index];
      while (tag) {
         length = 0;
         LengthXMLString(tag, &length);
         size += length;
         tag = tag->Next;
      }
   }
   else {
      size = 0;
      LengthXMLString(Self->Tags[Args->Index], &size);
   }

   /*** Allocate a buffer large enough to hold the XML information ***/

   size++;
   if (AllocMemory(size, MEM_STRING|MEM_CALLER, &buffer, NULL) IS ERR_Okay) {
      offset = 0;
      tag = Self->Tags[Args->Index];
      while (tag) {
         BuildXMLString(tag, buffer, Args->Flags, &offset);

         if (Args->Flags & XMF_INCLUDESIBLINGS) tag = tag->Next;
         else break;
      }

      Args->Result = buffer;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, MT_GetXMLString, ERR_AllocMemory));
}

/*****************************************************************************
** XML: Init
*/

static ERROR XML_Init(struct XML *Self, APTR Void)
{
   /* If an error occurred during compilation when the Location or
   ** Statement fields were set, return that error code.
   */

   if (Self->prvCompileError) return(Self->prvCompileError);

   /* NOTE: We do not fail if nothing has been loaded into the XML object,
   ** the developer may be creating an XML data structure from scratch,
   ** or could intend to send us information later.
   */

   return(ERR_Okay);
}

/*****************************************************************************
** XML: NewObject
*/

static ERROR XML_NewObject(struct XML *Self, APTR Void)
{
   if (AllocMemory(sizeof(APTR), MEM_DATA, &Self->Tags, NULL) IS ERR_Okay) {
      Self->prvCompileError = ERR_Okay;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_NewObject, ERR_AllocMemory));
}

/*****************************************************************************

###METHOD###
Name:    RemoveXMLTag
Synonym: RemoveTag
Short:   Removes tag(s) from the XML structure.

###DESCRIPTION###
<p>The RemoveTag method is used to remove one or more tags from an XML
structure. Any child tags that are discovered will automatically be removed
as a consequence of using this method, in order to maintain a valid XML
structure.</p>

<p>This method is capable of deleting multiple tags if you set the Total
parameter to a value greater than 1.  If you do this, each consecutive tag
following the first will also be removed from the XML structure. This is
useful for mass delete operations.</p>

<p>After using this method, you must assume that all tag addresses have been
changed due to the rearrangement of the XML structure.  Thus if you have
obtained pointers to various parts of the XML structure then you should
proceed to consider them invalid and re-establish them.</p>

###INPUT###
LONG/Index: The index of the tag that you want to remove.
LONG/Total: The total number of neighbouring tags that should also be deleted.  A value of one or less will remove just the indicated tag and its children.

###END###

*****************************************************************************/

static ERROR XML_RemoveXMLTag(struct XML *Self, struct mtRemoveXMLTag *Args)
{
   struct XMLTag *last_tag, *tag;
   LONG count, index, actual_count, i;

   if ((!Args) OR (Args->Index < 0) OR (Args->Index > Self->TagCount)) return(ObjectError(Self, MT_RemoveXMLTag, ERR_Args));

   index = Args->Index;
   count = Args->Total;
   if (count < 1) count = 1;

   #ifdef DEBUG
   DPrintF("RemoveTag()","[XML] Tag: %d, Total: %d", index, count);
   #endif

   /* Determine what the last valid tag would be when the tags
   ** are removed (may be NULL if there would be no further tags
   ** in this section of the XML hierarchy).
   */

   last_tag = Self->Tags[index]->Next;
   for (i=0; (i < count) AND (last_tag); i++) {
      last_tag = Self->Tags[index+i]->Next;
   }

   /* If we ran out of tags, update the count variable so that it is
   ** accurate.  Note that this figure counts for top-level tags only
   ** and does not include child tags.
   */

   if (i < count) count = i + 1;

   /* Calculate the total number of tags that we are going to remove,
   ** including child tags.
   */

   actual_count = 0;
   tag = Self->Tags[index];
   for (i=0; i < count; i++) {
      actual_count++;
      if (tag->Child) TagCount(tag->Child, &actual_count);
      tag = tag->Next;
   }

   /*** Fix up the address pointers of neighbouring tags ***/

   if (index > 0) {
      if (Self->Tags[index-1]->Child IS Self->Tags[index]) {
         /* If the tag to be removed is the child of a parent tag, fix
         ** up the parent's child pointer so that it points to the next
         ** child in the list.
         */
         Self->Tags[index-1]->Child = last_tag;
      }
      else {
         if (Self->Tags[index]->Prev) {
            /* Fix up the Next pointer of the previous neighbouring tag */
            Self->Tags[index]->Prev->Next = last_tag;
         }
      }
   }

   if (last_tag) last_tag->Prev = Self->Tags[index]->Prev;

   /*** Remove the tags from the array ***/

   for (i=index; (i < index + actual_count); i++) {
      if (Self->Tags[i]) FreeMemory(Self->Tags[i]);
      Self->Tags[i] = NULL;
   }

   /*** Clean up the hole that we have left in the tag list array ***/

   CopyMemory(Self->Tags + index + actual_count,
              Self->Tags + index,
              sizeof(APTR) * (Self->TagCount - (index + actual_count)));

   Self->TagCount -= actual_count; /* Subtract the total number of tags that were removed */
   Self->Tags[Self->TagCount] = NULL; /* Terminate the array */

   /*** Repair index numbers **/

   for (i=0; i < Self->TagCount; i++) {
      Self->Tags[i]->Index = i;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Reset
Short: Clears the information held in an XML object.
###END###

*****************************************************************************/

static ERROR XML_Reset(struct XML *Self, APTR Void)
{
   Action(AC_Clear, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  SaveToObject
Short: Saves XML data to a data storage object.
###END###

*****************************************************************************/

static ERROR XML_SaveToObject(struct XML *Self, struct acSaveToObject *Args)
{
   struct acWrite write;
   struct XMLTag *tag;
   STRING buffer;
   LONG offset, length, size;

   if ((!Args) OR (!Args->DestinationID)) return(ObjectError(Self, ERH_SaveToObject, ERR_Args));

   if (Self->TagCount <= 0) return(ERR_Okay);

   /*** Calculate the size of the buffer required to save the XML information ***/

   size = 0;
   tag = Self->Tags[Self->CurrentTag];
   while (tag) {
      length = 0;
      LengthXMLString(tag, &length);
      size += length;
      tag = tag->Next;
   }

   /*** Allocate a buffer large enough to hold the XML information ***/

   size++;
   if (AllocMemory(size, MEM_STRING, &buffer, NULL) IS ERR_Okay) {
      offset = 0;
      tag = Self->Tags[Self->CurrentTag];
      while (tag) {
         BuildXMLString(tag, buffer, NULL, &offset);
         tag = tag->Next;
      }

      /*** Write the data to the destination object ***/

      write.Buffer = buffer;
      write.Length = size;
      if (ActionMsg(AC_Write, Args->DestinationID, &write) != ERR_Okay) {
         FreeMemory(buffer);
         return(ERR_Write);
      }

      FreeMemory(buffer);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_SaveToObject, ERR_AllocMemory));
}

/*****************************************************************************

###METHOD###
Name:    SetXMLAttribute
Synonym: SetAttribute
Short:   Adds and updates XML attributes.

###DESCRIPTION###
<p>This method is used to update and add attributes to existing XML tags.
You need to supply the address of the tag that the attribute is for and the
Index number of the attribute that you are going to update.  If you are
adding a new attribute to the tag, you must use an Index of -1.</p>

<p>The data for the attribute is defined in the Name and Value parameters.
You can set a Value of NULL if no value is to be associated with the
attribute.  If both Name and Value are NULL, an error will be returned.

<p>Please note that the attribute at index 0 declares the name of the tag and
should not normally be accompanied with a value declaration.  However,
if the tag represents content within its parent, then the Name must be set
to NULL and the Value string will determine the content.</p>

<p>WARNING: The address of the original tag will be invalidated as a result
of calling this method.  You will need to regain the tag address if you
are using tag address pointers after calling this routine.</p>

###INPUT###
LONG/Index: Reference to the XML tag that you want to update.
LONG/Attrib: The index number of the attribute that is to be updated (set to -1 if this is a new attribute).
STRING/Name: String containing the new name for the attribute.  If NULL, the name will not be changed.
STRING/Value: String containing the new value for the attribute.

###RESULT###
ERR_Okay: The attribute was set.
ERR_Args: The required arguments were not specified.
ERR_OutOfRange: The Index or Attrib value is out of range.

###END###

*****************************************************************************/

static ERROR XML_SetXMLAttribute(struct XML *Self, struct mtSetXMLAttribute *Args)
{
   struct XMLTag *tag, *newtag;
   struct XMLAttribute *attrib;
   STRING buffer, name, value;
   LONG i, j, namelen, valuelen, pos, attribsize;

   if (!Args) return(ObjectError(Self, MT_SetXMLAttribute, ERR_Args));

   if ((Args->Index < 0) OR (Args->Index >= Self->TagCount)) {
      DPrintF("@SetAttrib()","[XML] Index out of range 0 <= %d < %d", Args->Index, Self->TagCount);
      return(ERR_OutOfRange);
   }

   #ifdef DEBUG
   DPrintF("SetAttrib()","[XML] Tag: %d, Attrib: %d, Name: %s, Value: %s", Args->Index, Args->Attrib, Args->Name, Args->Value);
   #endif

   if (Args->Attrib >= 0) {
      /* This routine deals with situations where the developer
      ** is setting an existing attribute.
      */

      tag = Self->Tags[Args->Index];

      if (Args->Attrib >= tag->TotalAttributes) {
         return(ObjectError(Self, MT_SetXMLAttribute, ERR_OutOfRange));
      }

      /*** Get the lengths of the name and value strings ***/

      name  = Args->Name;
      value = Args->Value;
      if ((!name) OR (!name[0])) name = tag->Attributes[Args->Attrib].Name;

      if (name) for (namelen=0; name[namelen]; namelen++);
      else namelen = 0;

      if (value) for (valuelen=0; value[valuelen]; valuelen++);
      else valuelen = 0;

      /*** Calculate the new size of our buffer ***/

      attrib = tag->Attributes;
      attribsize = tag->AttribSize;

      /* Remove the existing name and value sizes for the
      ** indexed attribute from the total.
      */

      if (attrib[Args->Attrib].Name) {
         for (i=0; attrib[Args->Attrib].Name[i]; i++);
         attribsize -= i + 1;
      }

      if (attrib[Args->Attrib].Value) {
         for (i=0; attrib[Args->Attrib].Value[i]; i++);
         attribsize -= i + 1;
      }

      /* Add the length of the new name and value strings to
      ** the overall size.
      */

      if (name)  attribsize += namelen + 1;
      if (value) attribsize += valuelen + 1;

      /* If the new buffer size is less or equal to the existing
      ** buffer's size, we can simply overwrite the existing attribute.
      ** Otherwise, we need to build a new buffer from scratch.
      */

      if (attribsize <= tag->AttribSize) {
         if (attrib[Args->Attrib].Name) buffer = attrib[Args->Attrib].Name;
         else if (attrib[Args->Attrib].Value) buffer = attrib[Args->Attrib].Value;
         else return(ObjectError(Self, MT_SetXMLAttribute, ERR_ObjectCorrupt));

         if (name) {
            for (i=0; i < namelen; i++) *buffer++ = name[i];
            *buffer++ = 0;
         }
         else attrib[Args->Attrib].Name = NULL;

         if (value) {
            attrib[Args->Attrib].Value = buffer;
            for (i=0; i < valuelen; i++) *buffer++ = value[i];
            *buffer++ = 0;
         }
         else attrib[Args->Attrib].Value = NULL;

         return(ERR_Okay);
      }
      else if (AllocMemory(sizeof(struct XMLTag) + Self->PrivateDataSize + (sizeof(struct XMLAttribute) * tag->TotalAttributes) + attribsize,
         MEM_DATA, &newtag, NULL) IS ERR_Okay) {

         /*** Copy the old tag details to the new tag ***/

         CopyMemory(tag, newtag, sizeof(struct XMLTag) + Self->PrivateDataSize);

         /*** Clean up the new tag and neighbouring tags ***/

         newtag->Private    = ((APTR)newtag) + sizeof(struct XMLTag);
         newtag->Attributes = ((APTR)newtag) + sizeof(struct XMLTag) + Self->PrivateDataSize;
         newtag->AttribSize = attribsize;
         if (newtag->Prev) newtag->Prev->Next = newtag;
         if (newtag->Next) newtag->Next->Prev = newtag;
         if ((newtag->Index > 0) AND (Self->Tags[newtag->Index-1]->Child IS tag)) Self->Tags[newtag->Index-1]->Child = newtag;

         /*** Recreate attribute names and values ***/

         buffer = ((APTR)newtag->Attributes) + (sizeof(struct XMLAttribute) * newtag->TotalAttributes);
         pos = 0;
         for (i=0; i < newtag->TotalAttributes; i++) {
            if (i IS Args->Attrib) {
               /*** Use the new name/value strings ***/
               if (name) {
                  newtag->Attributes[i].Name = buffer+pos;
                  for (j=0; j < namelen; j++) buffer[pos++] = name[j];
                  buffer[pos++] = 0;
               }

               if (value) {
                  newtag->Attributes[i].Value = buffer+pos;
                  for (j=0; j < valuelen; j++) buffer[pos++] = value[j];
                  buffer[pos++] = 0;
               }
            }
            else {
               /*** Use the old name/value strings ***/
               if (attrib[i].Name) {
                  newtag->Attributes[i].Name = buffer+pos;
                  for (j=0; attrib[i].Name[j]; j++) buffer[pos++] = attrib[i].Name[j];
                  buffer[pos++] = 0;
               }

               if (attrib[i].Value) {
                  newtag->Attributes[i].Value = buffer+pos;
                  for (j=0; attrib[i].Value[j]; j++) buffer[pos++] = attrib[i].Value[j];
                  buffer[pos++] = 0;
               }
            }
         }

         FreeMemory(tag);
         Self->Tags[Args->Index] = newtag;
         return(ERR_Okay);
      }
      else return(ObjectError(Self, MT_SetXMLAttribute, ERR_AllocMemory));
   }
   else {
      /*** Add a new attribute ***/

      DPrintF("@SetAttrib:","[XML] Adding attributes is not supported yet.");
      return(ERR_NoSupport);
   }
}

/*****************************************************************************

###METHOD###
Name:  SortXML
Short: Sorts XML tags to your specifications.

###DESCRIPTION###
<p>The SortXML method is used to sort a block of XML tags to your specifications.
You are required to pass a pointer to the tag that contains each tag item that
you want to sort.  If the tags lie at the root level, specify NULL in the Root
argument.  You need to provide an array that specifies sort instructions in the
List argument.  The array must be terminated with a series of NULL fields.
The array structure is as follows:</p>

<pre>
   struct XMLSort {
      STRING TagName;
      STRING Attribute;
      STRING Flags;
   }
</pre>

<p>The TagName determines the tag that should be used for sorting each node
(which means that you can sort on information held in child tags).  You may set
this to "Default" if the sort information is to be gathered from the root tags
as the sort algorithm scours the chain.</p>

<p>The Attribute field defines what attribute contains the sort data in each
XML tag.  This should be set to NULL if you want to sort on the content that
is found in the XML tag.</p>

<p>The Flags field supports the following options: XSF_DESC forces a
descending sort; XSF_CHECKSORT tells the algorithm to check for a 'sort'
attribute in each analysed tag and if found, the algorithm should use that
as the sort data instead of that indicated in the Attribute field.  If the
XSF_REPORTSORTED flag is used in the first array entry, the function will report
a code of ERR_NothingDone in the event that a rearrangement of the list was
not necessary.</p>

###INPUT###
LONG/Index: Sort everything <i>under</i> the specified tag, or -1 to sort the entire top level.
*XMLSort/List: Pointer to an array of sorting information.

###RESULT###
ERR_Okay: The XML object was successfully sorted.
ERR_Args: Invalid arguments were specified.
ERR_AllocMemory: A memory allocation error occurred.
ERR_NothingDone: The XML array was already sorted to your specifications.   Dependent on XSF_REPORTSORTED.

###END###

*****************************************************************************/

inline static BYTE SortNames(STRING Name1, STRING Name2)
{
   UBYTE char1, char2;

   while ((*Name1) AND (*Name2)) {
      char1 = *Name1;
      char2 = *Name2;
      if ((char1 >= 'A') AND (char1 <= 'Z')) char1 = char1 - 'A' + 'a';
      if ((char2 >= 'A') AND (char2 <= 'Z')) char2 = char2 - 'A' + 'a';

      if ((char1 > char2) OR ((Name1[1]) AND (!Name2[1]))) return(1); /* Name1 is greater */
      else if (char1 < char2) return(-1); /* Name1 is lesser */

      Name1++;
      Name2++;
   }

   return(0);
}

struct ListSort {
   struct XMLTag *Tag;  /* Pointer to the XML tag */
   UBYTE String[80];    /* Sort data */
};

static void SiftDown(struct ListSort **lookup, LONG i, LONG heapsize)
{
   struct ListSort *temp;
   LONG largest, left, right;

   largest = i;
   do {
      i = largest;
      left	= (i << 1) + 1;
      right	= left + 1;

      if (left < heapsize){
         if (SortNames(lookup[largest]->String, lookup[left]->String) > 0) largest = left;

         if (right < heapsize) {
            if (SortNames(lookup[largest]->String, lookup[right]->String) > 0) largest = right;
         }
      }

      if (largest != i) {
         temp = lookup[i];
         lookup[i] = lookup[largest];
         lookup[largest] = temp;
      }
   } while (largest != i);
}

static void SiftUp(struct ListSort **lookup, LONG i, LONG heapsize)
{
   struct ListSort *temp;
   LONG largest, left, right;

   largest = i;
   do {
      i = largest;
      left	= (i << 1) + 1;
      right	= left + 1;

      if (left < heapsize){
         if (SortNames(lookup[largest]->String, lookup[left]->String) < 0) largest = left;

         if (right < heapsize) {
            if (SortNames(lookup[largest]->String, lookup[right]->String) < 0) largest = right;
         }
      }

      if (largest != i) {
         temp = lookup[i];
         lookup[i] = lookup[largest];
         lookup[largest] = temp;
      }
   } while (largest != i);
}

static ERROR XML_SortXML(struct XML *Self, struct mtSortXML *Args)
{
   struct ListSort *list, **lookup, *temp;
   struct XMLTag *tag, **clone_array, *tmp, *child;
   LONG root_total, i, j, pos, index, insert_index, sort_total, tagcount;
   BYTE found, rearranged;

   if ((!Args) OR (!Args->List)) return(ObjectError(Self, MT_SortXML, ERR_Args));

   if (!Self->Tags) return(ERR_Okay);

   if (Args->Index < 0) tag = Self->Tags[0];
   else tag = Self->Tags[Args->Index]->Child;

   if (!tag) return(ERR_Okay);

   insert_index = tag->Index;

   /*** Count the number of root-tags to be sorted (does not include child tags) ***/

   for (tmp=tag, root_total=0; tmp; tmp=tmp->Next, root_total++);

   if (root_total < 2) return(ERR_Okay);

   sort_total = 0;
   TagCount(Self->Tags[insert_index], &sort_total);

   #ifdef DEBUG
   DPrintF("SortXML()","[%d] Index: %d, Tag: %s, Attrib: %s, Root-Total: %d, Sort-Total: %d of %d", Self->Head.UniqueID, insert_index, Args->List[0].TagName, Args->List[0].Attribute, root_total, sort_total, Self->TagCount);
   #endif

   /*** Allocate an array to store the sort results ***/

   if (AllocMemory(sizeof(struct ListSort) * root_total, MEM_NOCLEAR, &list, NULL) != ERR_Okay) {
      return(ERR_AllocMemory);
   }

   if (AllocMemory(sizeof(APTR) * root_total, MEM_NOCLEAR, &lookup, NULL) != ERR_Okay) {
      FreeMemory(list);
      return(ERR_AllocMemory);
   }

   for (i=0; i < root_total; i++) lookup[i] = list + i;

   /* Copy the tags into the sorting array (we just need to extract the
   ** data string for each tag).
   */

   for (index=0; (tag) AND (index < root_total); index++) {
      pos = 0;
      list[index].String[0] = 0;
      list[index].Tag = 0;
      for (i=0; (Args->List[i].TagName) AND (pos < sizeof(list[0].String)); i++) {

         found = FALSE;

         if (StrCompare("Default", Args->List[i].TagName, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            tmp = tag;
            found = TRUE;
         }
         else {
            if (tag->Next) j = tag->Next->Index;
            else j = Self->TagCount;
            for (tmp=tag; (tmp) AND (tmp->Index < j); tmp=Self->Tags[tmp->Index+1]) {
               if (StrCompare(Args->List[i].TagName, tmp->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  found = TRUE;
                  break;
               }
            }
         }

         if (found IS TRUE) {
            if (Args->List[i].Flags & XSF_CHECKSORT) {
               /*** Scan for a 'sort' attribute in the XML tag ***/
               for (j=0; j < tmp->TotalAttributes; j++) {
                  if (StrCompare("sort", tmp->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     pos += StrCopy(tmp->Attributes[j].Value, list[index].String+pos, sizeof(list[index].String)-pos);
                     found = FALSE; /* Turning off this flag will skip normal attribute extraction */
                     break;
                  }
               }
            }

            if (found IS TRUE) {
               if (!Args->List[i].Attribute) {
                  /*** Use content as the sort data ***/
                  for (child=tmp->Child; (child) AND (pos < sizeof(list[index].String)); child=child->Next) {
                     if (!child->Attributes[0].Name) {
                        pos += StrCopy(child->Attributes[0].Value, list[index].String+pos, sizeof(list[index].String)-pos);
                     }
                  }
               }
               else {
                  /*** Extract the sort data from the specified tag attribute ***/
                  for (j=0; j < tmp->TotalAttributes; j++) {
                     if (StrCompare(tmp->Attributes[j].Name, Args->List[i].Attribute, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        pos += StrCopy(tmp->Attributes[j].Value, list[index].String+pos, sizeof(list[index].String)-pos);
                        break;
                     }
                  }
               }
            }
         }

         /*** Each string in the sort list is separated with a byte value of 0x01 ***/

         if (pos < sizeof(list[0].String)-1) list[index].String[pos++] = 0x01;
      }

      if (pos >= sizeof(list[0].String)) pos = sizeof(list[0].String) - 1;
      list[index].String[pos] = 0;
      list[index].Tag = tag;

      tag = tag->Next;
   }

#ifdef SHELL_SORT
   /* Shell sort.  Similar to bubble sort but much faster
   ** because it can copy records over larger distances.
   */

   BYTE test;
   LONG h;

   h = 1;
   while (h < root_total / 9) h = 3 * h + 1;

   rearranged = FALSE;
   if (Args->List[0].Flags & XSF_DESC) {
      for (; h > 0; h /= 3) {
         for (i=h; i < root_total; i++) {
            temp = lookup[i];
            for (j=i; (j >= h) AND (SortNames(lookup[j - h]->String, temp->String) < 0); j -= h) {
               lookup[j] = lookup[j - h];
               rearranged = TRUE;
            }
            lookup[j] = temp;
         }
      }
   }
   else {
      for (; h > 0; h /= 3) {
         for (i=h; i < root_total; i++) {
            temp = lookup[i];
            for (j=i; (j >= h) AND (SortNames(lookup[j - h]->String, temp->String) > 0); j -= h) {
               lookup[j] = lookup[j - h];
               rearranged = TRUE;
            }
            lookup[j] = temp;
         }
      }
   }

#else
   /* Binary heap sort.  An extremely fast sorting algorithm, but assumes
   ** that the list has been rearranged on completion.
   */

   LONG heapsize;

   if (Args->List[0].Flags & XSF_DESC) {
      for (i=root_total/2; i >= 0; i--) SiftDown(lookup, i, root_total);

      heapsize = root_total;
      for (i=heapsize; i > 0; i--) {
         temp = lookup[0];
         lookup[0] = lookup[i-1];
         lookup[i-1] = temp;
         SiftDown(lookup, 0, --heapsize);
      }
   }
   else {
      for (i=root_total/2; i >= 0; i--) SiftUp(lookup, i, root_total);

      heapsize = root_total;
      for (i=heapsize; i > 0; i--) {
         temp = lookup[0];
         lookup[0] = lookup[i-1];
         lookup[i-1] = temp;
         SiftUp(lookup, 0, --heapsize);
      }
   }

   rearranged = TRUE;

#endif

   /*** Return if no sorting was required ***/

   if (rearranged IS FALSE) {
      FreeMemory(list);
      FreeMemory(lookup);
      if (Args->List[0].Flags & XSF_REPORTSORTED) return(ERR_NothingDone);
      else return(ERR_Okay);
   }

   /*** Clone the original tag array which will act as the target ***/

   if (AllocMemory(sizeof(APTR) * (Self->TagCount + 1), MEM_NOCLEAR, &clone_array, NULL) != ERR_Okay) {
      FreeMemory(list);
      FreeMemory(lookup);
      return(ObjectError(Self, MT_SortXML, ERR_Memory));
   }

   CopyMemory(Self->Tags, clone_array, sizeof(APTR) * (Self->TagCount + 1));

   index = insert_index;
   for (i=0; i < root_total; i++) {
      /*** Determine the total number of tags in the block to be copied ***/
      tagcount = 1;
      if (lookup[i]->Tag->Child) TagCount(lookup[i]->Tag->Child, &tagcount);

      CopyMemory(Self->Tags + lookup[i]->Tag->Index, clone_array + index, sizeof(APTR) * tagcount);

      if (i < root_total - 1) clone_array[index]->Next = lookup[i+1]->Tag;
      else clone_array[index]->Next = NULL; // The last tag must be null-terminated

      index += tagcount;
   }

   FreeMemory(Self->Tags);
   Self->Tags = clone_array;

   /*** Reset index numbers within the sorted range ***/

   for (i=insert_index; i < insert_index + sort_total; i++) {
      Self->Tags[i]->Index = i;
   }

   /*** Reset prev pointers for top-most tags in the sorted range ***/

   Self->Tags[0]->Prev = NULL;
   for (tag=Self->Tags[insert_index]; tag; tag=tag->Next) {
      if (tag->Next) tag->Next->Prev = tag;
   }

   FreeMemory(list);
   FreeMemory(lookup);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   CurrentTag
Short:  Determines the index of the main tag to use when building XML strings.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>When using any XML function that creates an XML string (e.g. SaveToObject),
the XML object will include the entire XML tree by default.  If you would like
to change this behaviour so that only a certain section of the tree is
included in the build, set the CurrentTag to the index of the tag that you
would like to start building from.  You can learn the index number of an
XMLTag by reading its Index field.  An index of zero will cover the entire
XML hierarchy.</p>

<p>The CurrentTag field currently affects the SaveToObject action and the
Statement field.</p>

###END###

****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags.
Type:   LONG
Prefix: XMF
Status: Read/Write

###DESCRIPTION###
<p>Special flags that apply to the XML class are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">INCLUDECOMMENTS&nbsp;&nbsp;</td><td>Normally comments will be stripped from XML statements during processing, but you may keep comment information by setting this flag.</td></tr>
<tr><td width="10%">LOWERCASE</td><td>This flag forces all attributes in each XML tag into lowercase characters.</td></tr>
<tr><td width="10%">STRIPCONTENT</td><td>Setting this flag will strip all content from an XML statement during processing.</td></tr>
<tr><td width="10%">UPPERCASE</td><td>This flag forces all attributes in each XML tag into uppercase characters.</td></tr>
<tr><td width="10%">WELLFORMED</td><td>By default, the XML class will accept badly structured XML data.  By setting this flag, all processed XML statements must be well formed (tags must balance) or a BadData error will be returned during processing.</td></tr>
</table>

###END###

****************************************************************************/

/*****************************************************************************

###FIELD###
Name:     Location
Synonyms: Src
Short:    Set this field if the XML statement originates from a file source.
Type:     STRING
Status:   Read/Write

###DESCRIPTION###
<p>If an XML statement is located in a text file, you may specify a pointer
to that file by setting this field.  If you set this field after
initialisation then the XML object will automatically load the file and
process it for you (this will clear out the previous set of XML data).</p>

###SEE ALSO###
Field: Statement

###END###

****************************************************************************/

static ERROR GET_Location(struct XML *Self, STRING *Value)
{
   if (Self->Location) { *Value = Self->Location; return(ERR_Okay); }
   else return(ERR_NoData);
}

static ERROR SET_Location(struct XML *Self, STRING Value)
{
   struct acRead read;
   STRING statement;
   OBJECTPTR file;
   LONG i;

   if (Self->Location) { FreeMemory(Self->Location); Self->Location = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->Location, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Location[i] = Value[i];
         Self->Location[i] = 0;

         if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
            FID_Location|TSTRING, Self->Location,
            FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
            TAGEND) IS ERR_Okay) {

            GetField(file, FID_TimeStamp, FT_LARGE, &Self->TimeStamp);

            if ((GetField(file, FID_Size, FT_LARGE, &Self->FileSize) IS ERR_Okay) AND (Self->FileSize > 0)) {
               if (AllocMemory(Self->FileSize+1, MEM_STRING|MEM_NOCLEAR, (APTR *)&statement, NULL) IS ERR_Okay) {
                  read.Buffer = statement;
                  read.Length = Self->FileSize;
                  if (Action(AC_Read, file, &read) IS ERR_Okay) {
                     statement[Self->FileSize] = 0;
                     SetFields(Self, FID_Statement|TSTRING, statement, TAGEND);
                  }
                  FreeMemory(statement);
               }
               else {
                  Action(AC_Free, file, NULL);
                  return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
               }
            }

            Action(AC_Free, file, NULL);
         }
         else {
            ObjectError(Self, ERH_SetField, ERR_CreateObject);
            return(ERR_Okay); /* Delay the return of the error (the Init action will do it) */
         }
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Statement
Short:  XML data is processed through this field.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>To send a string to an XML object, set the Statement field with a
pointer to the XML formatted data.  If you set this field after initialisation
then the XML object will respond by clearing out any existing data and
processing the new information that you have passed to it.</p>

<p>Be warned that setting this field with an invalid statement will result
in an empty XML object.</p>

<p>If you read the Statement field, a memory block will be returned that
contains a string based version of the XML object's data.  By default all
tags will be included in the statement, but you can start from a specific tag
by setting the <a href="#CurrentTag">CurrentTag</a> field.  You will need to
free the memory block returned by this field when you are done with it.</p>

###END###

*****************************************************************************/

static ERROR GET_Statement(struct XML *Self, STRING *Value)
{
   struct XMLTag *tag;
   STRING buffer;
   LONG offset, length, size;

   if ((Self->TagCount <= 0) OR (!Self->Tags)) return(ERR_FieldNotSet);

   /*** Calculate the size of the buffer required to save the XML information ***/

   size = 0;
   tag = Self->Tags[Self->CurrentTag];
   while (tag) {
      length = 0;
      LengthXMLString(tag, &length);
      size += length;
      tag = tag->Next;
   }

   /*** Allocate a buffer large enough to hold the XML information ***/

   size++;
   if (AllocMemory(size, MEM_STRING, &buffer, NULL) IS ERR_Okay) {
      offset = 0;
      tag = Self->Tags[Self->CurrentTag];
      while (tag) {
         BuildXMLString(tag, buffer, NULL, &offset);
         tag = tag->Next;
      }

      *Value = buffer;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_SaveToObject, ERR_AllocMemory));
}

static ERROR SET_Statement(struct XML *Self, STRING Value)
{
   if ((Value) AND (*Value)) {
      Self->prvCompileError = ConvertXML(Self, Value);

      if (Self->Head.ObjectFlags & NF_INITIALISED) return(Self->prvCompileError);
      else return(ERR_Okay); /* Delay the return of the error (the Init action will do it) */
   }
   else return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TagCount
Short:  Reflects the total number of tags in the XML array.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>This field can be read to learn the amount of tags in an XML
<a href="#Tags">Tags</a> array.  Zero is returned if there are no tags
in the XML object.</p>

###SEE ALSO###
Field: Tags

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Tags
Short:  Points to an array of tags loaded into an XML object.
Type:   struct XMLTag **
Status: Read

###DESCRIPTION###
<p>After initialising data to an XML object, you can read the processed
information through the Tags field.  The Tags field is a null terminated
array of all the XML tags that were retrieved from the original XML data.
The XML tags are represented in an XMLTag structure, which is arranged as
follows:</p>

<pre>
   struct XMLTag {
      LONG Index;                       Position within the XML array.
      struct XMLTag *Child;             Reference to further child tags.
      struct XMLTag *Prev;              Reference to the previous tag.
      struct XMLTag *Next;              Reference to the next tag.
      struct XMLAttribute *Attributes;  Attributes linked to the tag.
      WORD TotalAttributes;             The total number of tag attributes.
      WORD Size;                        Size of the structure in bytes.
   };

   struct XMLAttribute {
      STRING Name;
      STRING Value;
   };
</pre>

<p>Each XMLTag will also have at least one attribute set in the Attributes
array (the first attribute reflects the tag name) and the exact number of
attributes is indicated by the TotalAttributes field.</p>

<p>There are two methods that you can use for reading the XML tags; the first
is to scan through the array using a <i>for</i> loop (this gives a 'flat' view
of the data).  The second is to read the first tag in the array and use
the Next, Prev and Child fields to recurse through the XML tags.  Please
note that while scanning through the data you may encounter XML tags that
have a Name setting of NULL in the first attribute.  These tags represent
XML content, which you will find in the Value string of the attribute
structure.</p>

###SEE ALSO###
Field: TagCount

###END###

*****************************************************************************/

static ERROR CountTags(struct XML *Self, STRING Text, STRING *Result)
{
   STRING str;
   ERROR error;

   if (*Text != '<') {
      DPrintF("@ConvertXML:","Malformed XML statement detected.");
      return(ERR_BadData);
   }
   else Text++;

   /*** Skip past the tag's attributes ***/

   str = Text;
   while ((*str) AND (*str != '>')) {
      if ((str[0] IS '/') AND (str[1] IS '>')) break;

      skipwhitespace(str);

      if ((!*str) OR (*str IS '>') OR ((*str IS '/') AND (str[1] IS '>')) OR (*str IS '=')) break;

      while ((*str > 0x20) AND (*str != '>') AND (*str != '=')) {
         if ((str[0] IS '/') AND (str[1] IS '>')) break;
         str++;
      }

      skipwhitespace(str);

      if (*str IS '=') {
         str++;
         skipwhitespace(str);
         if (*str IS '"') {
            str++;
            while ((*str) AND (*str != '"')) str++;
            if (*str IS '"') str++;
         }
         else while ((*str > 0x20) AND (*str != '>')) {
            if ((str[0] IS '/') AND (str[1] IS '>')) break;
            str++;
         }
      }
   }

   if ((*str IS '>') AND (*Text != '!') AND (*Text != '?')) {
      /* We have found an unclosed tag.  This means that we must
      ** extract the content within it and handle any child tags.
      */

      str++;
      skipwhitespace(str);
      if (*str != '<') {
         while ((*str) AND (*str != '<')) str++;
         Self->TagCount++; /* A content tag will be required */
      }

      while ((str[0] IS '<') AND (str[1] != '/')) {
         if ((error = CountTags(Self, str, &str)) IS ERR_Okay) {
            skipwhitespace(str);
            if (*str != '<') {
               while ((*str) AND (*str != '<')) str++;
               Self->TagCount++; /* An embedded content tag will be required */
            }
         }
         else return(error);
      }

      /*** There should be a closing tag - skip past it ***/

      if ((str[0] IS '<') AND (str[1] IS '/')) {
         while ((*str) AND (*str != '>')) str++;
      }

      if (*str IS '>') str++;
   }
   else if ((str[0] IS '/') AND (str[1] IS '>')) str += 2;

   *Result = str;
   Self->TagCount++;
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ConvertXML()
*/

static ERROR ConvertXML(struct XML *Self, STRING Text)
{
   struct XMLTag *prevtag;
   STRING str;
   LONG i, j, tagindex;

   if ((!Self) OR (!Text)) return(ERR_Args);

   Self->prvBalance = NULL;

   /*** Kill any existing tags in this XML object ***/

   if (Self->Tags) {
      for (i=0; i < Self->TagCount; i++) {
         if (Self->Tags[i]) FreeMemory(Self->Tags[i]);
      }

      FreeMemory(Self->Tags);
      Self->Tags = NULL;
      Self->TagCount = 0;
   }

   /*** Strip out all comments ***/

   while ((i = StrSearch("<!--", Text, TRUE)) != -1) {
      if ((j = StrSearch("-->", Text+i, TRUE) + i) != -1) {
         j += 3;
         while (Text[j]) Text[i++] = Text[j++];
         Text[i] = 0;
      }
      else {
         DPrintF("@XML:","Detected malformed comment (missing --> terminator).");
         return(ERR_BadData);
      }
   }

   /* Perform a count of the total amount of tags specified (includes
   ** opening tags only, no closing tags).
   */

   Self->TagCount = 0;
   for (str=Text; (*str) AND (*str != '<'); str++);
   while ((str[0] IS '<') AND (str[1] != '/')) {
      if (CountTags(Self, str, &str) != ERR_Okay) {
         DPrintF("@XML:","Aborting XML interpretation process.");
         return(ERR_Failed);
      }
      while ((*str) AND (*str != '<')) str++;
   }

   if (Self->TagCount < 1) {
      DPrintF("@ConvertXML()","There are no valid tags in the XML statement.");
      return(ERR_BadData);
   }

   #ifdef DEBUG
   DPrintF("XML:","Detected %d raw and content based tags.", Self->TagCount);
   #endif

   /*** Allocate an array to hold all of the XML tags ***/

   if (AllocMemory(sizeof(APTR) * (Self->TagCount + 1), MEM_DATA, &Self->Tags, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Function, ERR_AllocMemory));
   }

   /* Extract the tag information.  This loop will extract the
   ** top-level tags.  The ExtractTag() function is recursive and
   ** will thus get the child tags.
   */

   tagindex = 0;
   prevtag = NULL;
   for (str=Text; (*str) AND (*str != '<'); str++);
   while ((str[0] IS '<') AND (str[1] != '/')) {
      i = tagindex; /* Remember the current tag index before ExtractTag() changes it */

      if (ExtractTag(Self, str, &str, &tagindex) != ERR_Okay) {
         DPrintF("@XML:","Aborting XML interpretation process.");
         return(ERR_Failed);
      }

      while ((*str) AND (*str != '<')) str++; /* Skip content/whitespace to get to the next tag */

      if (prevtag) prevtag->Next = Self->Tags[i];

      prevtag = Self->Tags[i];
   }

   #ifdef DEBUG
   DPrintF("XML:","%d tags successfully extracted.", Self->TagCount);
   #endif

   /* If the XML statement contained errors such as unclosed tags,
   ** the tag count may greater than the actual number of tags
   ** loaded.  This routine checks that the expected tag count
   ** matches what was extracted.
   */

   for (tagindex=0; tagindex < Self->TagCount; tagindex++) {
      if (!Self->Tags[tagindex]) {
         DPrintF("Tag:","Error - %d tags expected, loaded %d.", Self->TagCount, tagindex);
         if (tagindex > 0) {
            Self->Tags[tagindex-1]->Next = NULL;
            Self->Tags[tagindex-1]->Child = NULL;
         }
         Self->Tags[tagindex] = 0;
         Self->TagCount = tagindex;
         break;
      }
   }

   /* If the WELLFORMED flag has been used, check that the tags balance.
   ** If they don't then return ERR_BadData.
   */

   if (Self->Flags & XMF_WELLFORMED) {
      if (Self->prvBalance != 0) {
         DPrintF("@XML:","The XML tags do not balance.");
         return(ERR_BadData);
      }
   }

   /*** Set the previous field pointer and index field ***/

   for (i=0; i < Self->TagCount; i++) {
      Self->Tags[i]->Index = i;
      if (Self->Tags[i]->Next) Self->Tags[i]->Next->Prev = Self->Tags[i];
   }

   /*** Upper/lowercase transformations ***/

   if (Self->Flags & XMF_UPPERCASE) {
      for (tagindex=0; tagindex < Self->TagCount; tagindex++) {
         for (j=0; j < Self->Tags[tagindex]->TotalAttributes; j++) {
            str = Self->Tags[tagindex]->Attributes[j].Name;
            while (*str) if ((*str >= 'a') AND (*str <= 'z')) *str = *str - 'a' + 'A';
            str = Self->Tags[tagindex]->Attributes[j].Value;
            while (*str) if ((*str >= 'a') AND (*str <= 'z')) *str = *str - 'a' + 'A';
         }
      }
   }
   else if (Self->Flags & XMF_LOWERCASE) {
      for (tagindex=0; tagindex < Self->TagCount; tagindex++) {
         for (j=0; j < Self->Tags[tagindex]->TotalAttributes; j++) {
            str = Self->Tags[tagindex]->Attributes[j].Name;
            while (*str) if ((*str >= 'A') AND (*str <= 'Z')) *str = *str - 'A' + 'a';
            str = Self->Tags[tagindex]->Attributes[j].Value;
            while (*str) if ((*str >= 'A') AND (*str <= 'Z')) *str = *str - 'A' + 'a';
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ExtractTag()
*/

static ERROR ExtractTag(struct XML *Self, STRING Text, STRING *Result, LONG *TagIndex)
{
   struct XMLTag *tag, *child;
   LONG index, pos, currentindex, attribsize;
   WORD totalattrib;
   STRING str, buffer;
   ERROR error;

   /*** Check that the tag index does not exceed the total number of calculated tags ***/

   if (*TagIndex >= Self->TagCount) {
      DPrintF("@XML:","Ran out of array space for tag extraction (expected %d tags).", *TagIndex);
      return(ERR_Failed);
   }

   if (*Text != '<') {
      DPrintF("@ConvertXML:","Malformed XML statement detected.");
      return(ERR_BadData);
   }
   else Text++;

   currentindex = *TagIndex;

   Self->prvBalance++;

   /*** Count the number of tag attributes ***/

   totalattrib = 0;
   attribsize = 0;
   str = Text;
   while ((*str) AND (*str != '>')) {
      if ((str[0] IS '/') AND (str[1] IS '>')) break;

      skipwhitespace(str);

      if ((*str IS NULL) OR (*str IS '>') OR ((*str IS '/') AND (str[1] IS '>'))) break;

      if (*str IS '=') return(ObjectError(Self, ERH_Function, ERR_BadData));

      while ((*str > 0x20) AND (*str != '>') AND (*str != '=')) {
         if ((str[0] IS '/') AND (str[1] IS '>')) break;
         str++; attribsize++;
      }
      attribsize++; /* string termination byte */

      skipwhitespace(str);

      if (*str IS '=') {
         str++;
         skipwhitespace(str);
         if (*str IS '"') {
            str++;
            while ((*str) AND (*str != '"')) { str++; attribsize++; }
            if (*str IS '"') str++;
         }
         else while ((*str > 0x20) AND (*str != '>')) {
            if ((str[0] IS '/') AND (str[1] IS '>')) break;
            str++; attribsize++;
         }

         attribsize++; /* string termination byte */
      }

      totalattrib++;
   }

   if (totalattrib <= 0) {
      DPrintF("@ConvertXML:","Failed to extract a tag from \"%.10s\".", Text);
      return(ERR_BadData);
   }

   if (AllocMemory(sizeof(struct XMLTag) + Self->PrivateDataSize + (sizeof(struct XMLAttribute) * totalattrib) + attribsize,
         MEM_DATA, &tag, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Function, ERR_AllocMemory));
   }

   tag->Private         = ((APTR)tag) + sizeof(struct XMLTag);
   tag->Attributes      = ((APTR)tag) + sizeof(struct XMLTag) + Self->PrivateDataSize;
   tag->TotalAttributes = totalattrib;
   tag->AttribSize      = attribsize;

   Self->Tags[*TagIndex] = tag;

   TagIndex[0] += 1;

   /*** Extract all attributes within the tag ***/

   buffer = ((APTR)tag->Attributes) + (sizeof(struct XMLAttribute) * tag->TotalAttributes);
   str    = Text;
   index  = 0;
   pos    = 0;
   skipwhitespace(str);
   while ((*str) AND (*str != '>')) {
      if ((str[0] IS '/') AND (str[1] IS '>')) break;

      if (*str IS '=') return(ObjectError(Self, ERH_Function, ERR_BadData));

      /*** Extract the name of the attribute ***/

      tag->Attributes[pos].Name = buffer;
      while ((*str > 0x20) AND (*str != '>') AND (*str != '=')) {
         if ((str[0] IS '/') AND (str[1] IS '>')) break;
         *buffer++ = *str++;
      }
      *buffer++ = 0;

      /*** Extract the attribute's value ***/

      skipwhitespace(str);

      if (*str IS '=') {
         str++;
         skipwhitespace(str);
         tag->Attributes[pos].Value = buffer;
         if (*str IS '"') {
            str++;
            while ((*str) AND (*str != '"')) *buffer++ = *str++;
            if (*str IS '"') str++;
         }
         else {
            while ((*str > 0x20) AND (*str != '>')) {
               if ((str[0] IS '/') AND (str[1] IS '>')) break;
               *buffer++ = *str++;
            }
         }

         *buffer++ = 0;
      }

      pos++;
      skipwhitespace(str);
   }

   if ((*str IS '>') AND (tag->Attributes[0].Name[0] != '!') AND (tag->Attributes[0].Name[0] != '?')) {
      /* We have reached the end of an unclosed tag.  This means that
      ** we must extract the content within it and handle any child tags.
      */

      str++;

      index = *TagIndex; /* Remember the current tag position */
      if ((error = ExtractContent(Self, str, &str, TagIndex)) IS ERR_Okay) {
         tag->Child = Self->Tags[index]; /* Point our tag to the extract content */
      }
      else if (error != ERR_NoData) return(error);

      child = tag->Child;
      while ((str[0] IS '<') AND (str[1] != '/')) {
         index = *TagIndex; /* Remember the current tag index before ExtractTag() changes it */

         if (ExtractTag(Self, str, &str, TagIndex) IS ERR_Okay) {

            if (!tag->Child) tag->Child = Self->Tags[index];
            if (child) child->Next = Self->Tags[index];
            child = Self->Tags[index];

            /*** Extract any new content caught inbetween tags ***/

            index = *TagIndex;
            if ((error = ExtractContent(Self, str, &str, TagIndex)) IS ERR_Okay) {
               child->Next = Self->Tags[index];
               child = Self->Tags[index];
            }
            else {
               if (error != ERR_NoData) return(error);
            }
         }
         else return(ERR_Failed);
      }

      /*** There should be a closing tag - skip past it ***/

      if ((str[0] IS '<') AND (str[1] IS '/')) {
         Self->prvBalance--;
         while ((*str) AND (*str != '>')) str++;
      }

      if (*str IS '>') str++;
   }
   else {
      if ((str[0] IS '/') AND (str[1] IS '>')) str += 2;
      Self->prvBalance--;
   }

   *Result = str;
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ExtractContent()
*/

static ERROR ExtractContent(struct XML *Self, STRING Text, STRING *Result, LONG *TagIndex)
{
   struct XMLTag *tag;
   STRING str, buffer;
   LONG i, len;

   str = Text;

   skipwhitespace(str);

   /* If the STRIPCONTENT flag is set, we simply skip over the content
   ** and return a NODATA error code.
   */

   if (Self->Flags & XMF_STRIPCONTENT) {
      while ((*str) AND (*str != '<')) str++;
      *Result = str;
      return(ERR_NoData);
   }

   /*** Count size of the content and skip carriage returns (^M) ***/

   for (i=0, len=0; (str[i]) AND (str[i] != '<'); i++) {
      if (str[i] != '\r') len++;
   }

   if (len > 0) {
      if (AllocMemory(sizeof(struct XMLTag) + Self->PrivateDataSize + sizeof(struct XMLAttribute) + len + 1,
            MEM_DATA, &tag, NULL) IS ERR_Okay) {

         Self->Tags[*TagIndex] = tag;
         tag->Private         = ((APTR)tag) + sizeof(struct XMLTag);
         tag->Attributes      = ((APTR)tag) + sizeof(struct XMLTag) + Self->PrivateDataSize;
         tag->TotalAttributes = 1;
         tag->AttribSize      = len + 1;

         buffer = ((APTR)tag->Attributes) + sizeof(struct XMLAttribute);

         tag->Attributes[0].Name  = NULL;
         tag->Attributes[0].Value = buffer;

         while ((*str) AND (*str != '<')) {
            if (*str != '\r') *buffer++ = *str++;
            else str++;
         }
         *buffer = 0;

         TagIndex[0] += 1;
         *Result = str;
         return(ERR_Okay);
      }
      else {
         while ((*str) AND (*str != '<')) str++; /* Skip content */
         *Result = str;
         return(ERR_AllocMemory);
      }
   }
   else {
      *Result = str;
      return(ERR_NoData);
   }
}

/*****************************************************************************
** Internal: BuildXMLString()
** Short:    Converts XML data into its equivalent string.
*/

struct XMLTag * BuildXMLString(struct XMLTag *Tag, STRING Buffer, LONG Flags, LONG *Offset)
{
   struct XMLTag *xmltag;
   LONG offset, i, j;

   offset = *Offset;

   /*** If the tag is a content string, copy out the data and return ***/

   if (!Tag->Attributes[0].Name) {
      if (Tag->Attributes[0].Value) {
         if (Flags & XMF_UPPERCASE) {
            for (i=0; Tag->Attributes[0].Value[i]; i++) {
               Buffer[offset] = Tag->Attributes[0].Value[i];
               if ((Buffer[offset] >= 'a') AND (Buffer[offset] <= 'z')) Buffer[offset] = Buffer[offset] - 'a' + 'A';
               offset++;
            }
         }
         else if (Flags & XMF_LOWERCASE) {
            for (i=0; Tag->Attributes[0].Value[i]; i++) {
               Buffer[offset] = Tag->Attributes[0].Value[i];
               if ((Buffer[offset] >= 'A') AND (Buffer[offset] <= 'Z')) Buffer[offset] = Buffer[offset] - 'A' + 'z';
               offset++;
            }
         }
         else offset += StrCopy(Tag->Attributes[0].Value, Buffer+offset, COPY_ALL);
         Buffer[offset] = 0;
         *Offset = offset;

         //if (Flags & XMF_TRANSLATE) StrTranslate(Buffer, NULL, NULL); Bad concept - buffer space is fixed.
      }
      return(Tag->Next);
   }

   /*** Output the attributes assigned to this tag ***/

   Buffer[offset++] = '<';

   for (i=0; i < Tag->TotalAttributes; i++) {
      for (j=0; Tag->Attributes[i].Name[j]; j++) Buffer[offset++] = Tag->Attributes[i].Name[j];
      if (Tag->Attributes[i].Value) {
         Buffer[offset++] = '=';
         Buffer[offset++] = '"';
         for (j=0; Tag->Attributes[i].Value[j]; j++) Buffer[offset++] = Tag->Attributes[i].Value[j];
         Buffer[offset++] = '"';
      }
      if (i + 1 < Tag->TotalAttributes) Buffer[offset++] = ' ';
   }

   if ((xmltag = Tag->Child)) {
      Buffer[offset++] = '>';

      *Offset = offset;
      while (xmltag) {
         xmltag = BuildXMLString(xmltag, Buffer, Flags, Offset);
      }

      offset = *Offset;
      Buffer[offset++] = '<';
      Buffer[offset++] = '/';
      for (i=0; Tag->Attributes[0].Name[i]; i++) Buffer[offset++] = Tag->Attributes[0].Name[i];
      Buffer[offset++] = '>';
      Buffer[offset++] = '\n';
   }
   else {
      Buffer[offset++] = '/';
      Buffer[offset++] = '>';
      Buffer[offset++] = '\n';
   }

   Buffer[offset] = 0;
   *Offset = offset;
   return(Tag->Next);
}

/*****************************************************************************
** Internal: LengthXMLString()
** Short:    Determines the amount of bytes that would be required to write
**           out an XML string.
*/

struct XMLTag * LengthXMLString(struct XMLTag *Tag, LONG *Length)
{
   struct XMLTag *xmltag;
   LONG length, i, j;

   length = *Length;

   /*** If the tag is a content string, copy out the data and return ***/

   if (!Tag->Attributes[0].Name) {
      if (Tag->Attributes[0].Value) {
         for (i=0; Tag->Attributes[0].Value[i]; i++) length++;
         *Length = length;
      }
      return(Tag->Next);
   }

   /*** Output the attributes assigned to this tag ***/

   length++; // <

   for (i=0; i < Tag->TotalAttributes; i++) {
      for (j=0; Tag->Attributes[i].Name[j]; j++) length++;
      if (Tag->Attributes[i].Value) {
         length += 2; // ="
         for (j=0; Tag->Attributes[i].Value[j]; j++) length++;
         length++; // "
      }
      if (i + 1 < Tag->TotalAttributes) length++; // Space
   }

   if ((xmltag = Tag->Child)) {
      length++; // >
      *Length = length;
      while (xmltag) xmltag = LengthXMLString(xmltag, Length);
      length = *Length;
      length += 2; // </
      for (i=0; Tag->Attributes[0].Name[i]; i++) length++;
      length+=2;    // >\n
   }
   else length += 3; // />\n

   *Length = length;
   return(Tag->Next);
}

/*****************************************************************************
** Internal: TagCount
**
** This function calls itself recursively to count all tags (including children)
** within a tag space.  ALL sibling tags are also included.
*/

static void TagCount(struct XMLTag *Tag, LONG *Count)
{
   while (Tag) {
      if (Tag->Child) TagCount(Tag->Child, Count);
      *Count = *Count + 1;
      Tag = Tag->Next;
   }
}
